"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/* eslint-disable no-underscore-dangle */
const generator_1 = require("@babel/generator");
const contains_partition_keys_1 = require("./contains-partition-keys");
const executor_1 = require("./executor");
const parser_1 = require("./parser"); // eslint-disable-line import/no-unresolved
exports.SyntaxError = parser_1.SyntaxError;
const transformer_1 = require("./transformer");
class Query {
    constructor(query) {
        this._query = query;
        this._code = null;
        this.ast = parser_1.parse(this._query.trim());
    }
    get code() {
        if (!this._code) {
            const jsAst = transformer_1.default(this.ast);
            const { code } = generator_1.default(jsAst);
            this._code = code;
        }
        return this._code;
    }
    exec(coll, { parameters, udf, maxItemCount, continuation, compositeIndexes } = {}) {
        const { code } = this;
        if (!code) {
            throw new Error("Missing code");
        }
        return executor_1.default(coll, {
            code,
            parameters,
            udf,
            maxItemCount,
            continuation,
            compositeIndexes
        });
    }
    containsPartitionKeys(paths) {
        return contains_partition_keys_1.default(this.ast, paths);
    }
}
exports.default = (query) => new Query(query);
